package condition

import (
	"encoding/json"
	"fmt"
	"reflect"
	"strings"

	"git.fsdpf.net/go/db"
	"git.fsdpf.net/go/req"
	"github.com/samber/lo"
	"github.com/spf13/cast"
)

type ConditionOperator string
type ConditionTokenType string

const (
	SQL    ConditionTokenType = "sql"
	FUNC                      = "func"
	PARAM                     = "param"
	STRING                    = "string"
)

const (
	IS_NULL     ConditionOperator = "IS NULL"
	IS_NOT_NULL                   = "IS NOT NULL"
	EQ                            = "="
	NE                            = "!="
	GT                            = ">"
	GE                            = ">="
	LT                            = "<"
	LE                            = "<="
	LIKE                          = "LIKE"
	NOT_LIKE                      = "NOT LIKE"
	IN                            = "IN"
	NOT_IN                        = "NOT IN"
	REGEXP                        = "REGEXP"
	NOT_REGEXP                    = "NOT REGEXP"
)

type ConditionTokenValue interface {
	GetParam(k string) req.GlobalParams
	GetGlobalParamsUser() req.User
}

type ConditionExpr struct {
	parent            *Condition
	operator          ConditionOperator
	field             string
	fieldResource     string
	fieldSqlFunc      string
	fieldSqlFuncParam string
	ignoreEmptyParma  bool
	tokenType         ConditionTokenType
	token             string
	matchPrefix       string // 匹配前缀
}

func (this ConditionExpr) GetOperator() ConditionOperator {
	return this.operator
}

func (this *ConditionExpr) SetMatchPrefix(s string) *ConditionExpr {
	this.matchPrefix = s
	return this
}

func (this ConditionExpr) GetField() string {
	return this.field
}

func (this ConditionExpr) GetFieldResource() string {
	return this.fieldResource
}

func (this *ConditionExpr) AppendTo(c *Condition) {
	this.parent = c
}

func (this ConditionExpr) ToSql(m ConditionTokenValue) db.Expression {
	first := "`" + this.fieldResource + "`.`" + this.field + "`"

	if strings.Contains(this.field, "->") {
		first = "`" + this.fieldResource + "`." + this.field + ""
	}

	value := this.GetTokenSqlValue(m)

	operator := ConditionOperator(strings.ToUpper(string(this.GetOperator())))

	if value == "" {
		// @todo return true
		// value = "''"
	}

	//
	secondary := ""
	switch operator {
	case IS_NULL:
	case IS_NOT_NULL:

		secondary = ""
	case EQ, NE,
		GT, GE,
		LT, LE,
		REGEXP, NOT_REGEXP:

		if this.GetTokenType() == SQL {
			secondary = value
		} else {
			secondary = "'" + strings.Trim(value, "'") + "'"
		}

	case LIKE, NOT_LIKE:
		secondary = "'%" + strings.Trim(value, "'") + "%'"
	case IN, NOT_IN:
		secondary = "(" + lo.Ternary(value == "", "''", value) + ")"
	}

	if this.fieldSqlFunc == "json_member_of" {
		if this.fieldSqlFuncParam == "" {
			return db.Raw(fmt.Sprintf("JSON_CONTAINS(%s, JSON_ARRAY(%s))", secondary, first))
			// return db.Raw(fmt.Sprintf("%s MEMBER OF(%s)", secondary, first))
		} else {
			return db.Raw(fmt.Sprintf("JSON_CONTAINS(%s->>'%s', JSON_ARRAY(%s))", secondary, this.fieldSqlFuncParam, first))
			// return db.Raw(fmt.Sprintf("%s MEMBER OF(%s->'%s')", secondary, first, this.FieldSqlFuncParam))
		}
	} else if this.fieldSqlFunc == "json_contains" {
		if this.fieldSqlFuncParam == "" {
			return db.Raw(fmt.Sprintf("JSON_CONTAINS(%s, JSON_ARRAY(%s))", first, secondary))
		} else {
			return db.Raw(fmt.Sprintf("JSON_CONTAINS(%s->>'%s', JSON_ARRAY(%s))", first, this.fieldSqlFuncParam, secondary))
		}
	} else if this.fieldSqlFunc != "" && this.fieldSqlFuncParam != "" {
		first = this.fieldSqlFunc + "(" + first + ", " + this.fieldSqlFuncParam + ")"
	} else if this.fieldSqlFunc != "" {
		first = this.fieldSqlFunc + "(" + first + ")"
	}

	return db.Raw(strings.Trim(first+" "+string(operator)+" "+secondary, " "))
}

func (this ConditionExpr) GetTokenName() string {
	return this.token
}

func (this ConditionExpr) GetTokenType() ConditionTokenType {
	return this.tokenType
}

func (this *ConditionExpr) GetTokenSqlValue(m ConditionTokenValue) string {
	if this.GetTokenType() == SQL {
		return this.token
	}

	rv := reflect.ValueOf(this.GetTokenValue(m))

	if rv.Kind() == reflect.Ptr {
		rv = reflect.Indirect(rv)
	}

	switch rv.Kind() {
	case reflect.Invalid:
		return ""
	case reflect.Slice:
		aStr := cast.ToStringSlice(rv.Interface())

		for i := 0; i < len(aStr); i++ {
			if aStr[i] != "" {
				aStr[i] = "'" + strings.Trim(db.MysqlRealEscapeString(aStr[i]), "'") + "'"
			}
		}

		// 强制使用 in
		if this.operator == EQ {
			this.operator = IN
		}

		return strings.Join(aStr, ", ")
	case reflect.Struct:
		b, _ := json.Marshal(rv.Interface())
		return fmt.Sprintf("'%s'", b)
	default:
		return db.MysqlRealEscapeString(fmt.Sprintf("%v", rv.Interface()))
	}
}

func (this ConditionExpr) GetTokenValue(m ConditionTokenValue) any {
	switch this.GetTokenType() {
	case PARAM:
		if this.matchPrefix != "" {
			return m.GetParam(fmt.Sprintf("%s.%s", this.matchPrefix, this.token)).Value()
		}
		return m.GetParam(this.token).Value()
	case STRING:
		return this.token
	case FUNC:
		switch this.token {
		case "UserID":
			return m.GetGlobalParamsUser().ID()
		case "UserUuid":
			return m.GetGlobalParamsUser().Uuid()
		case "UserRolesUuid":
			return m.GetGlobalParamsUser().Roles()
		case "UserPlatform":
			return m.GetGlobalParamsUser().Runtime().Platform()
		case "UserSaaS":
			return m.GetGlobalParamsUser().Runtime().SaaS()
		}
	default:
		return nil
	}
	return nil
}

func (this ConditionExpr) IsIgnoreEmptyParma(m ConditionTokenValue) bool {
	if !this.ignoreEmptyParma {
		return false
	}

	if this.tokenType != PARAM {
		return false
	}

	param := this.GetTokenValue(m)

	if param == "" || param == nil {
		return true
	}

	return false
}

func NewExpr(rResource, rField, token string, operator ConditionOperator, tType ConditionTokenType, ignoreEmptyParma bool, fn, fnParam string) *ConditionExpr {
	return &ConditionExpr{
		operator:          operator,
		field:             rField,
		fieldResource:     rResource,
		fieldSqlFunc:      fn,
		fieldSqlFuncParam: fnParam,
		ignoreEmptyParma:  ignoreEmptyParma,
		tokenType:         tType,
		token:             string(token),
	}
}
